#!/usr/bin/env python


import httplib
import random
import re

from common import Common


class OptionCheck(Common):
    def __set_http_headers(self):
        """
        Extract random User-Agent HTTP header from a file and
        return into a dictionary
        """

        self.log("fetching random http user agent header from file '%s'" % self.args.headers_file_name)

        try:
            fd = open(self.args.headers_file_name)
        except:
            warn_msg  = "unable to read headers file "
            warn_msg += "'%s'" % self.args.headers_file_name
            self.warn(warn_msg)
            self.args.http_headers = {"User-Agent": "sqlmap/0.2 (http://sqlmap.sourceforge.net)"}
            return

        count = 0
        user_agents = []

        while True:
            line = fd.readline()
            if not line:
                break
            user_agents.append(line)
            count += 1

        fd.close()
        user_agent = user_agents[random.randint(1, count-1)]
        user_agent = user_agent.replace('\n', '').replace('\r', '')

        self.args.http_headers["User-Agent"] = user_agent

        self.log("fetched random http user agent header from file '%s'" % self.args.headers_file_name)


    def check(self, shellArgs):
        """
        Perform checks on some command line parameter values and
        set HTTP headers to be used for all HTTP requests in
        self.hash_page() function.
        """

        self.args = shellArgs
        self.args.parameters = {}

        # Perform checks on url (-u) parameter value
        url, parameters = self.args.url.split('?')
        for element in parameters.split('&'):
                elem = element.split('=')
                if len(elem) == 2:
                    parameter = elem[0]
                    if not self.args.url_parameter or parameter == self.args.url_parameter:
                        value = elem[1]
                        if value:
                            self.args.parameters[parameter] = value

        if not re.search('^http[s]*://[\w\.\-\_]+', self.args.url, re.I):
            port = re.search('^[\w\.\-\_]+:([\d]+)', self.args.url, re.I)
            if port:
                if int(port.groups()[0]) == 443:
                    self.args.url = 'https://' + self.args.url
                else:
                    self.args.url = 'http://' + self.args.url
            else:
                self.args.url = 'http://' + self.args.url

        link = re.search('^http[s]*://([\w\.\-\_]+)', self.args.url, re.I)
        transport = re.search('^(http[s]*)://[\w\.\-\_]+', self.args.url, re.I)
        port = re.search('^http[s]*://[\w\.\-\_]+:([\d]+)', self.args.url, re.I)

        if link:
            self.args.link = link.groups()[0]

        if transport:
            self.args.transport = transport.groups()[0]
        elif port:
            if int(port.groups()[0]) == 443:
                self.args.transport = 'https'
            else:
                self.args.transport = 'http'
        else:
            self.args.transport = 'http'

        if port:
            self.args.port = int(port.groups()[0])
        elif self.args.transport == 'http':
            self.args.port = 80
        else:
            self.args.port = 443

        if self.args.transport == 'http':
            self.args.http = httplib.HTTPConnection
        else:
            self.args.http = httplib.HTTPSConnection

        # Perform checks on headers file name (-f) parameter value
        self.args.http_headers = {}

        if self.args.headers_file_name:
            self.__set_http_headers()
        else:
            self.args.http_headers["User-Agent"] = "sqlmap/0.2 (http://sqlmap.sourceforge.net)"

        if self.args.cookie:
            self.args.http_headers["Connection"] = "Keep-Alive"
            self.args.http_headers["Cookie"] = self.args.cookie

        # Perform checks on method (--method) parameter value
        if self.args.http_method:
            self.args.http_method = self.args.http_method.upper()

            if self.args.http_method not in ("GET", "POST"):
                warn_msg  = "'%s' " % self.args.http_method
                warn_msg += "is an unsupported HTTP method, "
                warn_msg += "setting to default value"
                self.warn(warn_msg)
                self.args.http_method = "GET"
        else:
            self.args.http_method = "GET"

        # Perform checks on proxy (--proxy) parameter value
        if self.args.proxy:
            if not re.search("[\w\.\-\_]+\:[\w\.\-\_]+", self.args.proxy, re.I):
                err_msg  = "ERROR: proxy value must be in "
                err_msg += "format 'proxy_address:port'"
                raise Exception, err_msg
            else:
                self.args.proxy = self.args.proxy.replace("http://", "")
                self.args.proxy = re.search("([\w\.\-\_\:]+)",
                                            self.args.proxy,
                                            re.I).group()

        return self.args
