#!/usr/bin/env python


import sys
import time

from optparse import OptionParser
from optparse import OptionError

from option import OptionCheck
from injection import InjectionCheck
from dump import SQLMapDump
from mysqlmap import MySqlMap
from postgresmap import PostgresMap


def show_banner():
    print """
    sqlmap() v0.2 coded by belch <daniele.bellucci@gmail.com>
                      and inquis <bernardo.damele@gmail.com>
    """


def parse_args():
    parser = OptionParser()

    try:
        parser.add_option("-u",
                          dest="url",
                          help="target url")

        parser.add_option("-p",
                          dest="url_parameter",
                          help="injectable parameter")

        parser.add_option("-f",
                          dest="headers_file_name",
                          help="load http headers from file")

        parser.add_option("--method",
                          dest="http_method",
                          help="http method, GET or POST (default: GET)")

        parser.add_option("--cookie",
                          dest="cookie",
                          help="http cookie header")

        parser.add_option("--proxy",
                          dest="proxy",
                          help="use a proxy to connect to the url")

        parser.add_option("-b", "--banner",
                          dest="get_banner",
                          action="store_true",
                          help="get database banner")

        parser.add_option("--current-db",
                          dest="get_current_db",
                          action="store_true",
                          help="get current database name")
        
        parser.add_option("--current-user",
                          dest="get_current_user",
                          action="store_true",
                          help="get current database user")

        parser.add_option("--users",
                          dest="get_users",
                          action="store_true",
                          help="get database users")

        parser.add_option("--dbs",
                          dest="get_dbs",
                          action="store_true",
                          help="get available databases")

        parser.add_option("--tables",
                          dest="get_tables",
                          action="store_true",
                          help="get available databases tables (optional: -D)")

        parser.add_option("--columns",
                          dest="get_columns",
                          action="store_true",
                          help="get table columns (require: -T optional: -D)")

        parser.add_option("--dump",
                          dest="dump_table",
                          action="store_true",
                          help="dump a database table (require: -T optional: -D)")

        parser.add_option("--file",
                          dest="filename",
                          help="read a specific file content")

        parser.add_option("-e",
                          dest="expression",
                          help="expression to evaluate")

        parser.add_option("-D", "--database",
                          dest="db",
                          help="database to enumerate")

        parser.add_option("-T", "--table",
                          dest="table",
                          help="table to enumerate")

        parser.add_option("-v", "--verbose",
                          dest="verbose",
                          action="store_true",
                          help="be verbose")

        (args, arg) = parser.parse_args()

        if not args.url:
            parser.error("missing url parameter (Hint: -h)")

        return args
    except (OptionError, TypeError), e:
        parser.error(e)
        sys.exit(1)


def get_dbms_handler(args):
    try:
        app = MySqlMap(args)
        if app.check_dbms():
            return app

        app = PostgresMap(args)
        if app.check_dbms():
            return app
    except:
        return None


def main():
    show_banner()
    print "[*] starting at: %s\n" % (time.strftime("%X"))

    app = None
    shellArgs = parse_args()

    try:
        optObj = OptionCheck()
        checkedArgs = optObj.check(shellArgs)

        injObj = InjectionCheck()
        injArgs = injObj.check(checkedArgs)

        if not injArgs.injparameter:
            raise Exception, "ERROR: all parameters are not injectable"

        app = get_dbms_handler(injArgs)

        if not app:
            raise Exception, "ERROR: it is not possible to fingerprint remote database"

        print "remote database: %s\n" % app.get_fingerprint()
        dumper = SQLMapDump()

        if injArgs.get_banner:
            dumper.string("banner",
                          app.get_banner())

        if injArgs.get_current_db:
            dumper.string("current database",
                          app.get_current_db())

        if injArgs.get_current_user:
            dumper.string("current user",
                          app.get_current_user())

        if injArgs.get_users:
            dumper.list("database users",
                        app.get_users())

        if injArgs.get_dbs:
            dumper.list("available databases",
                        app.get_dbs())

        if injArgs.get_tables:
            dumper.db_tables(app.get_tables())

        if injArgs.get_columns:
            dumper.table_columns(app.get_columns())

        if injArgs.dump_table:
            dumper.table_values(app.dump_table())

        if injArgs.filename:
            dumper.string(injArgs.filename,
                          app.get_file(injArgs.filename))

        if injArgs.expression:
            dumper.string(injArgs.expression,
                          app.get_expression(injArgs.expression))
    except Exception, e:
        if app == None:
            print "\n%s\n" % str(e)
        else:
            app.warn(str(e))
        sys.exit(1)

    print "[*] shutting down at: %s\n" % (time.strftime("%X"))


if __name__ == "__main__":
    main()
